#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include<stdarg.h>
#include<time.h>
#include<ctype.h>
#include "bludi.h"

#define VPRAVO  0
#define VLEVO   1
#define DOLU    2
#define NAHORU  3
#define DOPREDU 4
#define DOZADU  5

int sirka, vyska, hloubka;
int SCstena[2], SCA[2], SCB[2], SCX[2], SCY[2], SCZ[2];
static char *pole, SCpolicko[2];
int *zdi_x, *zdi_y, *zdi_z, zdi_SC[2];

static struct prac{
  int x, y, z;
  char smer;
} *pracpole;
static int pracdelka, pracmax;

void warning(char *s, ...)
{
  va_list vl;

  fprintf(stderr, "warning: ");
  va_start(vl, s);
  vfprintf(stderr, s, vl);
  va_end(vl);
  fprintf(stderr, "\n");
}

void error(char *s, ...)
{
  va_list vl;

  fprintf(stderr, "error: ");
  va_start(vl, s);
  vfprintf(stderr, s, vl);
  va_end(vl);
  fprintf(stderr, "\n");
  exit(1);
}

static char *pole_sour(int x, int y, int z)
{
  int i;
  if(x<0 || y<0 || z<0 || x>=sirka || y>=vyska || z>=hloubka){
    for(i=0; i<2; i++) if(x==SCX[i] && y==SCY[i] && z==SCZ[i])
			 return &SCpolicko[i];
    return NULL;
  }
  return &pole[x+y*sirka+z*sirka*vyska];
}

char opacsmer(char smer)
{return smer/2*2+(1-smer%2);}

int *zed_sour(int x, int y, int z, char smer)
{
  int i;
  int *zdi, a,b,c, maxa,maxb,maxc;

  if(smer < 0 || smer > 5) return NULL;
  if(x<0 || y<0 || z<0 || x>=sirka || y>=vyska || z>=hloubka){
    for(i=0; i<2; i++)
      if(x==SCX[i] && y==SCY[i] && z==SCZ[i] &&
	 smer == opacsmer(SCstena[i]))
	return &zdi_SC[i];
    return NULL;
  }
  switch(smer/2){
  case 0:
    zdi = zdi_x; a=y; b=z; c=x; maxa=vyska; maxb=hloubka; maxc=sirka;
    break;
  case 1:
    zdi = zdi_y; a=x; b=z; c=y; maxa=sirka; maxb=hloubka; maxc=vyska;
    break;
  case 2:
    zdi = zdi_z; a=x; b=y; c=z; maxa=sirka; maxb=vyska; maxc=hloubka;
  }

  maxc--;
  if(smer%2) c--;
  if(c<0 || c >= maxc){
    for(i=0; i<2; i++)
      if(a == SCA[i] && b == SCB[i] && smer == SCstena[i])
	return &zdi_SC[i];
    return NULL;
  }
  return &zdi[a+b*maxa+c*maxa*maxb];
}

void posunsour(int *x, int *y, int *z, char smer)
{
  int *c;

  c = NULL;
  switch(smer/2){
  case 0: c=x; break;
  case 1: c=y; break;
  case 2: c=z; break;
  }
  if(!c){
    warning("neznamy smer %d", smer);
    return;
  }

  if(smer%2) (*c)--;
  else (*c)++;
}

void vypissmer(char smer)
{
  switch(smer){
  case VPRAVO:  printf("VPRAVO\n"); break;
  case VLEVO:   printf("VLEVO\n"); break;
  case DOLU:    printf("DOLU\n"); break;
  case NAHORU:  printf("NAHORU\n"); break;
  case DOPREDU: printf("DOPREDU\n"); break;
  case DOZADU:  printf("DOZADU\n"); break;
  default: warning("neznamy smer");
  }
}

static void init()
{
  int i, delka;

  delka = sirka*vyska*hloubka;
  pole = (char *)malloc(delka*sizeof(char));
  for(i=0; i<delka; i++) pole[i] = 0;

  delka = (sirka-1)*vyska*hloubka;
  zdi_x = (int *)malloc(delka*sizeof(int));

  delka = sirka*(vyska-1)*hloubka;
  zdi_y = (int *)malloc(delka*sizeof(int));

  delka = sirka*vyska*(hloubka-1);
  zdi_z = (int *)malloc(delka*sizeof(int));

  pracmax = (sirka-1)*vyska*hloubka+sirka*(vyska-1)*hloubka
    +sirka*vyska*(hloubka-1)+2;
  pracpole = (struct prac *)malloc(pracmax*sizeof(struct prac));

  for(i=0; i<2; i++){
    if(SCstena[i] < 2){
      SCY[i]=SCA[i];
      SCZ[i]=SCB[i];
      if(SCstena[i]==1) SCX[i] = -1;
      else SCX[i] = sirka;
    }
    else if(SCstena[i] < 4){
      SCX[i]=SCA[i];
      SCZ[i]=SCB[i];
      if(SCstena[i]==3) SCY[i] = -1;
      else SCY[i] = vyska;
    }
    else{
      SCX[i]=SCA[i];
      SCY[i]=SCB[i];
      if(SCstena[i]==5) SCZ[i] = -1;
      else SCZ[i] = hloubka;
    }

    pracpole[i].x = SCX[i];
    pracpole[i].y = SCY[i];
    pracpole[i].z = SCZ[i];
    pracpole[i].smer = opacsmer(SCstena[i]);
    zdi_SC[i] = i;
    SCpolicko[i] = 2*i-1;
  }
  pracdelka = 2;
}

/*
static void vypisprac()
{
  int i;
  for(i=0; i<pracdelka; i++){
    printf("%d:  %d, %d, %d,  ", i,pracpole[i].x,pracpole[i].y,pracpole[i].z);
    switch(pracpole[i].smer){
    case VPRAVO: printf("VPRAVO\n");break;
    case VLEVO: printf("VLEVO\n");break;
    case DOLU: printf("DOLU\n");break;
    case NAHORU: printf("NAHORU\n");break;
    case DOPREDU: printf("DOPREDU\n");break;
    case DOZADU: printf("DOZADU\n");break;
    }
  }
}
*/

static void pridejprac(int x, int y, int z, char smer)
{
  if(pracdelka >= pracmax){
    error("Bludi capacity exceeded!");
  }
  *zed_sour(x, y, z, smer) = pracdelka;
  pracpole[pracdelka].x = x;
  pracpole[pracdelka].y = y;
  pracpole[pracdelka].z = z;
  pracpole[pracdelka].smer = smer;
  pracdelka++;
}

static void odstranprac(int i, char zbyde)
{
  pracdelka--;
  *zed_sour(pracpole[pracdelka].x, pracpole[pracdelka].y,
	    pracpole[pracdelka].z, pracpole[pracdelka].smer) = i;
  *zed_sour(pracpole[i].x, pracpole[i].y, pracpole[i].z, 
	    pracpole[i].smer) = zbyde;
  pracpole[i] = pracpole[pracdelka];
}

static int *freqpole;
static int pocetslep;
static int delkabludi;
static int minzatacek;

static void prolomzed(int x, int y, int z, char smer)
{
  int i,delka,poslsmer;
  int x2, y2, z2, smer2;
  char *pol;

  *zed_sour(x, y, z, smer) = 0;

  if(*pole_sour(x,y,z) > 0){
    posunsour(&x, &y, &z, smer);
    smer = opacsmer(smer);
  }

  if(*pole_sour(x,y,z) < -6) pocetslep++;
  x2=x; y2=y; z2=z;
  posunsour(&x2,&y2,&z2,smer);
  if(*pole_sour(x2,y2,z2) > 6) pocetslep++;  

  delka = sirka*vyska*hloubka;
  for(i=0; i<delka; i++){
    if(pole[i] < 0) pole[i] *= -1;
    if(pole[i] >= 7) pole[i] -= 7;
    else pole[i]--;
    if(pole[i] >= 6 || pole[i] < 0) warning("zvlastni cislo %d na policku %d", i,pole[i]);
  }

  minzatacek = 0;
  delkabludi = 1;

  x2=x; y2=y; z2=z;
  poslsmer = opacsmer(smer);
  while(x2 != SCX[0] || y2 != SCY[0] || z2 != SCZ[0]){
    pol = pole_sour(x2,y2,z2);
    if(!pol){
      warning("nenasel jsem cestu ke startu");
      return;
    }
    smer2 = *pol;
    *pol = 6+opacsmer(poslsmer)+smer2*6;

    if(poslsmer != smer2) minzatacek++;
    delkabludi++;

    posunsour(&x2, &y2, &z2, smer2);
    poslsmer = smer2;
 }

  x2=x; y2=y; z2=z;
  posunsour(&x2, &y2, &z2, smer);
  poslsmer = smer;
  while(x2 != SCX[1] || y2 != SCY[1] || z2 != SCZ[1]){
    pol = pole_sour(x2,y2,z2);
    if(!pol){
      warning("nenasel jsem cestu k cili");
      return;
    }
    smer2 = *pol;
    *pol = 6+smer2+opacsmer(poslsmer)*6;

    if(poslsmer != smer2) minzatacek++;
    delkabludi++;

    posunsour(&x2, &y2, &z2, smer2);
    poslsmer = smer2;
  }
}

static void prolomSC()
{
  int x, y, z, i, j;
  char old, new;

  for(i=0; i<2; i++)
    if(zdi_SC[i]){
      prolomzed(SCX[i], SCY[i], SCZ[i], opacsmer(SCstena[i]));
      return;
    }

  i=0;
  j=1;
  for(;;){
    for(x=0; x<sirka; x++)
      for(y=0; y<vyska; y++)
	for(z=0; z<hloubka; z++){
	  old = new;
	  new = *pole_sour(x, y, z);
	  if(z > 0 && old*new < 0){
	    i += j;
	    if(!i){
	      prolomzed(x, y, z, DOZADU);
	      return;
	    }
	  }
	}
    for(x=0; x<sirka; x++)
      for(z=0; z<hloubka; z++)
	for(y=0; y<vyska; y++){
	  old = new;
	  new = *pole_sour(x, y, z);
	  if(y > 0 && old*new < 0){
	    i += j;
	    if(!i){
	      prolomzed(x, y, z, NAHORU);
	      return;
	    }
	  }
	}
    for(y=0; y<vyska; y++)
      for(z=0; z<hloubka; z++)
	for(x=0; x<sirka; x++){
	  old = new;
	  new = *pole_sour(x, y, z);
	  if(x > 0 && old*new < 0){
	    i += j;
	    if(!i){
	      prolomzed(x, y, z, VLEVO);
	      return;
	    }
	  }
	}

    if(i == 0){
      warning("nenalezena hranice");
      return;
    }
    if(j == -1){
      warning("nedosazen konec");
      return;
    }
    i = rand()%i+1;
    j = -1;
  }
}

static void generuj()
{
  int i, x, y, z, smer, x2, y2, z2;
  char *pol1, *pol2;

  pocetslep = 0;

  while(pracdelka){
    i = rand()%pracdelka;
    x = pracpole[i].x;
    y = pracpole[i].y;
    z = pracpole[i].z;
    smer = pracpole[i].smer;
    odstranprac(i, 0);

    pol2 = pole_sour(x, y, z);
    if(!pol2) error("pol2 == NULL");

    posunsour(&x, &y, &z, smer);
    smer = opacsmer(smer);

    pol1 = pole_sour(x, y, z);
    if(!pol1) error("pol1 == NULL");
    if(*pol2 > 0){
      *pol1 = smer+1;
      if(*pol2 <= 6) *pol2 += 6;
      else pocetslep++;
    }
    else{
      *pol1 = -smer-1;
      if(*pol2 >= -6) *pol2 -= 6;
      else pocetslep++;
    }

    for(i=0; i<6; i++){
      if(i == smer) continue;
      x2 = x; y2 = y; z2 = z;
      posunsour(&x2, &y2, &z2, i);
      pol2 = pole_sour(x2, y2, z2);
      if(pol2){
	if(*pol2 == 0) pridejprac(x, y, z, i);
	else odstranprac(*zed_sour(x, y, z, i), 1);
      }
    }
  }
}

void vytvorbludi()
{
  int i, x,y,z;

  if(sirka <=0 || vyska <= 0 || hloubka <= 0)
    error("spatne rozmery bludiste: %dx%dx%d", sirka, vyska, hloubka);
  for(i=0; i<2; i++){
    if(SCstena[i] < 0 || SCstena[i] > 5)
      error("spatne cislo steny startu nebo cile, musi byt 0,1,2,3,4,5");
    if(SCstena[i] < 2){x=0, y=SCA[i], z=SCB[i];}
    else if(SCstena[i] < 4){x=SCA[i], y=0, z=SCB[i];}
    else {x=SCA[i], y=SCB[i], z=0;}
    if(x<0 || y<0 || z<0 || x>=sirka || y>=vyska || z>=hloubka)
      error("spatne souradnice startu nebo cile");
  }
  if(SCstena[0] == SCstena[1] && SCA[0] == SCA[1] && SCB[0] == SCB[1])
    error("start = cil");


  init();
  generuj();
  prolomSC();

  free(pracpole);
}

char jezed(int x, int y, int z, char smer)
{
  int *zed;

  zed = zed_sour(x,y,z, smer);
  if(!zed){
    if(x>=0 && y>=0 && z>=0 && x<sirka && y<vyska && z<hloubka) return 1;
    posunsour(&x, &y, &z, smer);
    if(x>=0 && y>=0 && z>=0 && x<sirka && y<vyska && z<hloubka) return 1;
    return 0;
  }
  return *zed;
}

void vykresli()
{
  int x, y, z;
  char c;

  for(z = hloubka-1; z>=0; z--){
    for(y=-1; y<vyska; y++){
      if(y>=0){
	for(x=-1; x<sirka; x++){
	  if(x>=0){
	    c=' ';
	    if(!jezed(x,y,z, DOPREDU)) c='O';
	    if(!jezed(x,y,z, DOZADU)){
	      if(c==' ') c='.';
	      else c = '*';
	    }
	    printf("%c", c);
	  }
	  if(jezed(x,y,z, VPRAVO)) printf("|");
	  else printf(" ");
	}
	printf("\n");
      }
      for(x=-1; x<sirka; x++){
	if(x>=0){
	  if(jezed(x,y,z, DOLU)) printf("-");
	  else printf(" ");
	}
	c = ' ';
	if(jezed(x,y,z, VPRAVO) || jezed(x,y+1,z, VPRAVO)) c = '|';
	if(jezed(x,y,z, DOLU) || jezed(x+1,y,z, DOLU)){
	  if(c == ' ') c = '-';
	  else c = '+';
	}
	printf("%c", c);
      }
      printf("\n");
    }
    if(z>0) printf("\n\n\n\n");
  }
}

void printhelp()
{
  printf("\n");
  printf("Parametry prikazove radky:\n");
  printf("  -h -? -help --help       tato napoveda\n");
  printf("\n");
  printf("  -s --size <A>    rozmery budou AxAxA (default 4)\n");
  printf("            <A>x<B>  rozmery budou AxBx1\n");
  printf("            <A>x<B>x<C>  rozmery budou AxBxC\n");
  printf("\n");
  printf("  -S --seed <string pro nahodny generator>\n");
  printf("\n");
  printf("  -l --level <cislo>\n");
  printf("     level->rozmery:  1->1x1x1, 2->2x1x1 3->2x2x1, 4->2x2x2,\n");
  printf("                      5->3x2x2, 10->4x4x4, 20->8x7x7, ...\n");
  printf("\n");
  printf("  -d --demo    automaticke projiti bludiste na zacatku\n");
  printf("\n");
  printf("  -f --fullscreen  rezim cele obrazovky\n");
  printf("                   (moznost prepnuti pri hre klavesou f)\n");
  printf("\n");
  printf("  -i --invert  prohozeni sipek nahoru a dolu\n");
  printf("                 (moznost prepnuti pri hre klavesou i)\n");
  printf("\n");
  printf("Ovladani:\n");
  printf("  otaceni sipkami, posun dopredu mezernikem\n");
  printf("Cil:\n");
  printf("  projit bludistem od modreho ctverecku k zelenemu nebo obracene.\n");
  printf("\n");
}

void paramerror()
{
  error("spatne parametry prikazove radky, zkuste parametr -h");
}

void nastavbludi(int argc, char **argv, char *fstart)
{
  int i, j;
  unsigned int seed;

  sirka = vyska = hloubka = 4;

  srand(time(NULL));

  for(j=1; j < argc; j++){
    if(!strcmp("-h", argv[j]) || !strcmp("-?", argv[j])  ||
       !strcmp("-help", argv[j]) || !strcmp("--help", argv[j])){
      printhelp();
      exit(0);
    }
    else if(!strcmp("-s", argv[j]) || !strcmp("--size", argv[j])){
      j++;
      if(j == argc) paramerror();
      sirka = vyska = hloubka = 0;
      if(!isdigit(argv[j][0])) paramerror();
      for(i=0; isdigit(argv[j][i]); i++) sirka = sirka*10 + argv[j][i]-'0';
      if(!argv[j][i]){
	hloubka = vyska = sirka;
	continue;
      }
      if(argv[j][i] != 'x') paramerror();
      if(!isdigit(argv[j][++i])) paramerror();
      for(; isdigit(argv[j][i]); i++) vyska = vyska*10 + argv[j][i]-'0';
      if(!argv[j][i]){
	hloubka = 1;
	continue;
      }
      if(argv[j][i] != 'x') paramerror();
      if(!isdigit(argv[j][++i])) paramerror();
      for(; isdigit(argv[j][i]); i++) hloubka = hloubka*10 + argv[j][i]-'0';
      if(argv[j][i]) paramerror();
    }
    else if(!strcmp("-l", argv[j]) || !strcmp("--level", argv[j])){
      j++;
      if(j == argc) paramerror();
      sirka = 0;
      if(!isdigit(argv[j][0])) paramerror();
      for(i=0; isdigit(argv[j][i]); i++) sirka = sirka*10 + argv[j][i]-'0';
      if(argv[j][i]) paramerror();

      hloubka = (sirka+2)/3;
      vyska = (sirka+3)/3;
      sirka = (sirka+4)/3;
    }
    else if(!strcmp("-S", argv[j]) || !strcmp("--seed", argv[j])){
      j++;
      if(j == argc) paramerror();
      seed=i=0;
      for(i=0; argv[j][i]; i++) seed = (seed*16+(argv[j][i]))%(3999);
      srand(seed);
    }
    else if(!strcmp("-d", argv[j]) || !strcmp("--demo", argv[j]))
      demo = 2;
    else if(!strcmp("-f", argv[j]) || !strcmp("--fullscreen", argv[j]))
      *fstart = 1;
    else if(!strcmp("-i", argv[j]) || !strcmp("--invert", argv[j]))
      invert = 1;
    else paramerror();
  }

  printf ("dim: %dx%dx%d\n", sirka, vyska, hloubka);

  SCstena[0] = NAHORU;
  SCA[0] = 0;
  SCB[0] = hloubka-1;
  SCstena[1] = DOLU;
  SCA[1] = sirka-1;
  SCB[1] = 0;
}

static int stav;
static int pokusu, maxfreq;
static int demosmer;
static int kstartuX, kstartuY, kstartuZ;

void analyzujbludi()
{
  int delka, i, x,y,z;

  delka = sirka*vyska*hloubka;

  freqpole = (int *)malloc(delka*sizeof(int));
  for(i = 0; i<delka; i++)
    freqpole[i] = 0;

  maxfreq = 0;
  stav = 0; pokusu = 0;

  kstartuX = SCX[0]+1; kstartuY = SCY[0]+1; kstartuZ = SCZ[0]+1;
  demosmer = 0;

  x = SCX[1]+1; y = SCY[1]+1; z = SCZ[1]+1;
  if(abs(x)+abs(y)+abs(z) < abs(kstartuX)+abs(kstartuY)+abs(kstartuZ)){
    kstartuX = x;
    kstartuY = y;
    kstartuZ = z;
    demosmer = 1;
  }
}

static int tahu, zatacek, navstivslep, poslsmer;

char analyzujtah(int x, int y, int z, char smer)
{
  int i, j;
  int x2, y2, z2;
  char vyhral;

  if(jezed(x,y,z,smer)) return 0;

  vyhral = 0;
  tahu++;
  if(smer != poslsmer){
    zatacek++;
    poslsmer = smer;
  }

  i = x+y*sirka+z*sirka*vyska;
  x2 = x; y2 = y; z2 = z;
  posunsour(&x2,&y2,&z2,smer);

  if(x2<0 || y2<0 || z2<0 || x2>=sirka || y2>=vyska || z2>=hloubka){
    if(stav < 0) return 1;
    if(x2 == SCX[stav] && y2 == SCY[stav] && z2 == SCZ[stav]) vyhral = 1;
    stav = -1;
    i = -1;
  }
  else{
    if(stav < 0){
      if(x == SCX[0] && y == SCY[0] && z == SCZ[0]) stav = 1;
      else stav = 0;
      tahu = 1; zatacek = 0; pokusu++;
    }
  }

  if(stav >= 0){
    j = x2+y2*sirka+z2*sirka*vyska;
    freqpole[j]++;
    if(freqpole[j] > maxfreq) maxfreq = freqpole[j];

    if(pole[j] >= 0 && pole[j] < 6){
      pole[j] -= 6;
      if(i < 0){
	warning("slepa ulicka na startu");
	return 1;
      }
      if(pole[i] >= -6 && pole[i] < 0) pole[i] -= 6;
      else navstivslep++;
    }
  }
  if(vyhral){
    printf("VYHRAL JSI\n");
    printf("Stacilo ti na to %d pokusu\n", pokusu);
    if(pocetslep == 0) i = 0;
    else i = 100*navstivslep/pocetslep;
    printf("Navstivil jsi %d z %d slepych ulicek, coz je %d%%\n",
	   navstivslep, pocetslep, i);
    printf("Nejnavstevovanejsi misto jsi prosel %d krat\n", maxfreq);
    printf("Na projiti bludiste ti stacilo %d posunuti a %d odboceni.\n",
	   tahu, zatacek);
    if(tahu == 0) i = 100;
    else i = 100*delkabludi/tahu;
    if(zatacek == 0) j = 100;
    else j = 100*minzatacek/zatacek;
    printf("Na projiti bludiste stacilo %d (%d%%) posunuti a %d (%d%%) odboceni\n",
	   delkabludi, i, minzatacek, j);
    return 2;
  }
  return 1;
}

void demotest()
{
  int x,y,z, smer;

  smer=x=y=z=-1;

  for(;;){
    smer = demotah(x,y,z);
    if(smer == -1) break;
    vypissmer(smer);
    posunsour(&x,&y,&z,smer);
  }
}

char demotah(int x, int y, int z)
{
  if(kstartuY > 0){
    kstartuY--;
    return DOLU;
  }
  if(kstartuX+x < kstartuZ+z)
    if(kstartuZ > 0){
      kstartuZ--;
      return DOPREDU;
    }
  if(kstartuX > 0){
    kstartuX--;
    return VPRAVO;
  }
  if(kstartuZ > 0){
    kstartuZ--;
    return DOPREDU;
  }
  if(x == SCX[1-demosmer] && y == SCY[1-demosmer] && z == SCZ[1-demosmer]) return -1;
  if(x == SCX[demosmer] && y == SCY[demosmer] && z == SCZ[demosmer])
    return opacsmer(SCstena[demosmer]);

  if(demosmer == 0) return (*pole_sour(x,y,z))%6;
  else return (*pole_sour(x,y,z))/6-1;
}

/*
int main(int argc, char ** argv)
{
  nastavbludi();
  vytvorbludi();
  vykresli();

  return 0;  
}
*/
